'''
    Read/write plain text files.

    This module can be replaced using python directly.
    But there are many times when you want to read and write
    simple text files without a lot of exceptions. In
    other words, if you try to read a file as lines and
    it doesn't exist, then you just get an empty list.

    Copyright 2008-2020 DeNova
    Last modified: 2020-10-20

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import shutil
from traceback import format_exc

from denova.python.log import get_log
from denova.python.utils import is_string

log = get_log()


def read(filename):
    '''Read the contents of a text file. Return the lines as a list.

       If the file doesn't exist, return empty list.
    '''

    try:
        inputFile = open(filename, 'rt')
        lines = inputFile.readlines()
        inputFile.close()
    except:   # 'bare except' because it catches more than "except Exception"
        log(f'Unable to read {filename}')
        log(format_exc())
        lines = []

    return lines


def write_line(filename, line, append=False):
    '''Write a line to a text file.

      Log any io errors and then raise another ioerrr.
    '''

    lines = [line]
    return write(filename, lines, append)


def write(filename, lines, append=False):
    '''Write the lines to a text file.

      Log any io errors and then raise another ioerrr.
    '''

    try:
        if append:
            method = 'at'
        else:
            method = 'wt'

        outputFile = open(filename, method)
        for line in lines:
            if isinstance(line, list):
                # line must be another list
                # let's assume there aren't any more nested lists
                for inner_line in line:
                    if is_string(l):
                        text = inner_line.decode()
                    else:
                        text = inner_line
                    outputFile.write(text)
            else:
                if is_string(line):
                    text = line
                else:
                    text = line.decode()
                outputFile.write(text)
        outputFile.close()

    except IOError:
        log(f'Unable to write {filename}')
        log(format_exc())
        raise IOError

    return lines


def backup_and_write(filename, lines, append=False):
    '''Backup the text file and then write the new content.

      Log any io errors and then raise another ioerrr.
    '''

    backup(filename)

    return write(filename, lines, append)


def backup(filename):
    '''Backup the text file.

      Log any io errors and then raise another ioerrr.
    '''

    try:
        shutil.copy2(filename, filename + '.bak')

    except IOError as io_error:
        (errno, strerror) = io_error.args
        log(f'Unable to backup {filename}')
        log(f"   ({errno:d}) {strerror}")
        raise IOError



'''
    Retry 404 response on another backend http server.

    If static resources such as css and js are served by a front end server,
    that static server must also serve the static resources of the other
    backend http server. See redirect404.py for an alternative. Or you may
    choose to bypass the front end server, at a cost in performance.

    The server is specified by settings.PROXY_404_SERVER.
    PROXY_404_SERVER is in the form:

        [USERNAME[:PASSWORD]@]HOST[:PORT]

    HOST can be a DNS name or IP address.
    PROXY_404_SERVER Examples::

        example.com
        example.com:8080
        username@example.com:8080
        username:password@example.com:8080
        127.0.0.1
        127.0.0.1:8080

    Copyright 2014-2020 DeNova
    Last modified: 2020-10-22

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import traceback
import urllib.request
from urllib.parse import urlsplit, urlunsplit

try:
    from django import http
    from django.conf import settings
    from django.utils.deprecation import MiddlewareMixin
except ModuleNotFoundError:
    import sys
    sys.exit('Django required')

from denova.python.log import get_log

log = get_log()


class Proxy404Middleware(MiddlewareMixin):

    def process_response(self, request, response):

        try:
            if response.status_code == 404:

                log('got 404')
                url = request.get_full_path()
                parts = urlsplit(url)

                if parts.scheme == '':
                    if request.is_secure():
                        scheme = 'https'
                    else:
                        scheme = 'http'
                else:
                    scheme = parts.scheme

                old_host = request.get_host()
                new_host = settings.PROXY_404_SERVER
                if new_host != old_host:

                    log(f'old host: {old_host}, new host: {new_host}') #DEBUG
                    new_url = urlunsplit([scheme, new_host,
                        parts.path, parts.query, parts.fragment])
                    log(f'url: {url}') #DEBUG
                    log('parts: %s' % repr(parts)) #DEBUG
                    log(f'new url: {new_url}') #DEBUG

                    try:
                        log(f'opening {new_url}') #DEBUG
                        stream = urllib.request.urlopen(new_url)
                        log(f'reading {new_url}') #DEBUG
                        try:
                            new_response = http.HttpResponse(stream.read())
                        finally:
                            stream.close()
                        if new_response.status_code == 404:
                            log('404 ' % new_url) #DEBUG
                        else:
                            log(f'got {new_url}') #DEBUG
                            response = new_response

                    except Exception:
                        # just log it and return the exsting 404
                        log(f'url: {url}')
                        log('parts: %s' % repr(parts))
                        log(f'new url: {new_url}')
                        log(traceback.format_exc())

        except Exception:
            log(traceback.format_exc())
            raise

        return response



'''
    Processes.

    We prefer to use command line programs over python standard modules
    because linux command line programs are more likely to have been
    thoroughly vetted.

    Copyright 2013-2020 DeNova
    Last modified: 2020-10-20

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import datetime
import os
import signal
import subprocess
import sys, threading
import time
import _thread
from contextlib import contextmanager
from traceback import format_exc

from denova.os.command import run
from denova.os.user import require_user
from denova.python.log import get_log
from denova.python.utils import stacktrace, version
from denova.python.times import now

if version < 3:
    sys.exit('requires python 3 or later')

log = get_log()
DEBUGGING = False

class TimedOutException(Exception):
    ''' Operation timed out exception. '''
    pass

class CallException(Exception):
    ''' call() exception. '''
    pass

def pid_from_port(port):
    ''' Find which pid opened the port.

        This can be much faster than trying to connect to a port.
        Calling socket.socket().connect() sometimes succeeds for every port.

        Returns None if none.

        >>> from denova.os.user import whoami
        >>> if whoami == 'root':
        ...     pids = pid_from_port(9988)
        ...     pids is None
        ... else: print(True)
        True
    '''

    # fuser requires --namespace before port, so can't use namespace='tcp'
    pids = pids_from_fuser('--namespace', 'tcp', port)
    if pids:
        assert len(pids) == 1, f'pids: {pids}'
        pid = pids[0]
        result = int(pid)
    else:
        result = None
    return result

def pids_from_file(path):
    ''' Get list of pids that have the file or dir open.

        Returns an empty list if no pids.

        >>> from denova.os.user import whoami
        >>> if whoami == 'root':
        ...     pids = pids_from_file('/tmp')
        ...     len(pids) == 0
        ... else: print(True)
        True
    '''

    return pids_from_fuser(path)

def pids_from_program(program):
    ''' Get list of pids for program.

        Returns empty list if none

        >>> from denova.os.user import whoami
        >>> if whoami == 'root':
        ...     pids = pids_from_program('python3')
        ...     len(pids) > 0
        ... else: print(True)
        True
    '''

    try:
        pid_strings = run('pidof', program).stdout.strip().split()
    except subprocess.CalledProcessError:
        pids = []
    else:
        pids = [int(pid) for pid in pid_strings]

    return pids

def program_from_pid(pid):
    ''' Find which program has the pid.

        The program is the full path of the running program. This may
        not match the program on the command line, if the program on
        the command line is a link.

        Returns None if none.

        >>> program = program_from_pid(9988)
        >>> program is None
        True
    '''

    # /proc/PID/exe is a link to the program
    try:
        program = os.readlink(f'/proc/{pid}/exe')
    except OSError:
        program = None

    log.debug(f'program_from_pid({pid}): {program}')
    return program

def program_from_port(port):
    ''' Find which program opened the port.

        See program_from_pid(pid).

        Returns None if none.

        >>> from denova.os.user import whoami
        >>> if whoami == 'root':
        ...     pid = program_from_port(9988)
        ...     pid is None
        ... else: print(True)
        True
    '''

    log.debug(f'port: {port}')
    pid = pid_from_port(port)
    log.debug(f'pid: {pid}')
    if pid:
        program = program_from_pid(pid)
    else:
        program = None

    return program

def programs_using_file(path):
    ''' Find which programs have the file or dir open.

        Returns None if none.

        >>> from denova.os.user import whoami
        >>> if whoami == 'root':
        ...     programs = programs_using_file('/tmp')
        ...     programs is None
        ... else: print(False)
        False
    '''

    programs = set()

    # fuser
    pids = pids_from_file(path)
    if pids:
        log.debug(f'pids using {path}: {pids}')

        for pid in pids:
            program = program_from_pid(pid)
            if program:
                programs.add(program)
            else:
                log.debug(f'no program from pid  {pid}')

    # lsof
    lines = run('lsof').stdout.strip().split('\n')
    for line in lines:
        fields = line.split()
        command = fields[0]
        command_path = fields[-1]
        if command_path == path or command_path.startswith(path + '/'):
            programs.add(command)

    if programs:
        programs = sorted(programs)
    else:
        programs = None

    return programs

def pids_from_fuser(*args):
    ''' Get list of pids using fuser.

        Returns Empty list if no pids.

        >>> from denova.os.user import whoami
        >>> if whoami == 'root':
        ...     pids = pids_from_fuser(*['--namespace', 'tcp', '9988'])
        ...     len(pids) == 0
        ... else: print(True)
        True
    '''

    require_user('root') # fuser requires root

    log.debug(f'pids_from_fuser(*args) args: {repr(args)}')
    try:
        ''' Example:
                # fuser --namespace tcp 9050
                9050/tcp:             3331

            WARNING: Very strangely, fuser sends the first part of the line to
            stderr, and the second part to stdout. The stdout splits on the
            spaces.

            fuser also returns an exit code of 1.
        '''
        fuser_out = run('fuser', *args, _ok_code=[0, 1])
    except:   # 'bare except' because it catches more than "except Exception"
        # log.debug(format_exc()) # DEBUG
        pids = []
    else:
        log.debug(f'pids_from_fuser() fuser_out: {repr(fuser_out)}')
        # only the pids g to stdout
        pid_strings = fuser_out.stdout.strip().split()
        pids = [int(pid) for pid in pid_strings]

    log.debug(f'pids_from_fuser() pid: {pids}')

    return pids

def is_program_running(program_name):
    '''
        Return True if program is running and not defunct.

        >>> is_program_running('python3')
        True
        >>> is_program_running('not.running')
        False
    '''

    return find_active_program(program_name) is not None

def get_pid(program_name):
    '''
        Return first matching pid if program is running and not defunct.

        >>> pid = get_pid('python3')
        >>> pid is not None
        True
        >>> type(pid)
        <class 'int'>
    '''

    pid = None

    line = find_active_program(program_name)
    if line is not None:
        index = line.find(' ')
        if index > 0:
            pid = int(line[:index])

    return pid

def get_path(program_name):
    '''
        Return first matching path if program is running and not defunct.

        >>> TARGET = 'python3'
        >>> path = get_path(TARGET)
        >>> path is not None
        True
        >>> type(path)
        <class 'str'>
    '''

    path = None

    line = find_active_program(program_name)
    if line is not None:
        parts = line.split()
        if len(parts) > 1:
            path = parts[1]

    return path

def find_active_program(program_name):
    '''
        Return the first matching raw line from "ps" if program is running.
        Ignores any program_name that is defunct.

        >>> line = find_active_program('python3')
        >>> line is not None
        True
    '''

    raw_line = None

    try:
        raw_lines = program_status(program_name)
        if raw_lines:
            raw_line = raw_lines[0]
    except:   # 'bare except' because it catches more than "except Exception"
        log(format_exc())
        raw_line = None

    return raw_line


def program_status(program_name):
    '''
        Returns a list of matching raw lines from "ps" if program is running.
        If no lines match, returns an empty list.
        Ignores any program_name that is defunct.

        >>> lines = program_status('python3')
        >>> lines == []
        False
    '''
    PS_ARGS = ['-eo', 'pid,args']

    lines = []
    try:
        raw = run('ps', *PS_ARGS)
        raw_lines = raw.stdout.strip().split('\n')

        for line in raw_lines:
            line = str(line).strip()
            # apparent the leading space matters
            if program_name in line and ' <defunct>' not in line:
                lines.append(line)
    except:   # 'bare except' because it catches more than "except Exception"
        log(format_exc())

    return lines

def wait(event, timeout=None, sleep_time=1, event_args=None, event_kwargs=None):
    ''' Wait for an event. Retries event until success or timeout.

        Default is to ignore exceptions except when there is a timeout.

        'event' is a function. event() succeeds if it does not raise an
        exception. Each call to event() continues until it succeeds or
        raises an exception. It is not interrupted if it times out.

        'timeout' can be in seconds as an int or float, or a
        datetime.timedelta, or a datetime.datetime. Default is None, which
        means no timeout. If the timeout deadline passes while event() is
        running, event() is not interrupted. If event() times out while
        running and does not succeed, wait() raises the exception from event().

        'sleep_time' is in seconds. Default is one.

        'event_args' is an list of positional args to event(). Default is None.
        'event_kwargs' is an dict of keyword args to event(). Default is None.

        Returns result from event() if no timeout, or if timeout returns last exception.
    '''

    def timed_out():
        return timeout and (now() >= deadline)

    if timeout:

        if isinstance(timeout, int) or isinstance(timeout, float):
            deadline = now() + datetime.timedelta(seconds=timeout)

        elif isinstance(timeout, datetime.timedelta):
            deadline = now() + timeout

        elif isinstance(timeout, datetime.datetime):
            deadline = timeout

        else:
            raise TypeError('timeout must be an int, float, datetime.timedelta, or datetime.datetime')

        log.debug(f'wait() timeout: {timeout}, deadline: {deadline}')

    if event_args is None:
        event_args = []
    if event_kwargs is None:
        event_kwargs = {}

    success = False
    while not success:
        try:
            result = event(*event_args, **event_kwargs)

        except KeyboardInterrupt:
            raise

        except:   # 'bare except' because it catches more than "except Exception"
            if timed_out():
                log.debug(f'wait() timed out with exception: {format_exc()}')
                raise
            else:
                log.debug('wait() ignored exception because not timed out:\n{}'.format(
                    format_exc()))

        else:
            success = True

        if not timed_out():
            time.sleep(sleep_time)

    return result

@contextmanager
def fork_child():
    ''' Context manager to run a forked child process.

        Subprocess returns a process result code of 0 on success,
        or -1 if the child block raises an exception.

        >>> parent_pid = os.getpid()
        >>> # if child process
        >>> if os.fork() == 0:
        ...     with fork_child():
        ...         # child process code goes here
        ...         assert parent_pid != os.getpid()
        >>> assert parent_pid == os.getpid()
    '''

    """ Including "if os.fork() == 0:" in the context manager results in::

            RuntimeError: generator didn't yield
    """

    # try to run in a new session
    try:
        os.setsid()
    except:   # 'bare except' because it catches more than "except Exception"
        pass

    # continue after the calling process ends
    signal.signal(signal.SIGHUP, signal.SIG_IGN)

    try:
        yield
    except:   # 'bare except' because it catches more than "except Exception"
        returncode = -1
    else:
        returncode = 0
    finally:
        os._exit(returncode)

def zombies():
    ''' Return list of zombies in the format::

            [(pid, parent_pid), ...]

        Sometimes when you stop a process, a zombie process lives.
        A process becomes a zombie when the process finishes running, but
        the parent process didn't wait for it. The fix is to make the
        parent wait for the subprocess. A workaround is to kill the parent
        process.
    '''

    pids = []

    stdout = run('ps', '-A', '--format', 'stat,pid,ppid').stdout
    for line in stdout.split('\n'):
        stat, pid, ppid = line.split()
        if 'Z' in stat.upper():
            pids.append((pid, ppid))

    return pids

def kill(pid):
    ''' Kill pid. '''

    # it takes multiple signals to kill reliably
    for sig in [signal.SIGTERM, signal.SIGHUP, signal.SIGKILL]:
        try:
            os.kill(pid, sig)
        except Exception:
            pass

def is_pid_active(pid):
    ''' Return True if pid is active. Else return False.

        >>> is_pid_active(os.getpid())
        True
    '''

    try:
        # this does not "kill" the process,
        # just checks to see if it's alive
        os.kill(pid, 0)
    except OSError:
        active = False
    else:
        active = True

    return active


if __name__ == "__main__":

    import doctest
    doctest.testmod()



'''
    Context processor to set all custom template variables.

    Copyright 2009-2020 DeNova
    Last modified: 2020-10-25

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

try:
    from django.conf import settings
except ModuleNotFoundError:
    import sys
    sys.exit('Django required')

from denova.net.browser import browser_types, is_primitive_browser
from denova.python.log import get_log

log = get_log()



def custom(request):
    ''' Django context processor to set all custom template variables. '''

    custom_context = {}

    #DEBUG log('request: %r' % request)

    #custom_context.update(is_live(request))
    custom_context.update(debug(request))
    custom_context.update(browser(request))
    custom_context.update(css_ok(request))
    custom_context.update(javascript_ok(request))
    custom_context.update(is_active_user(request))

    return custom_context

def browser(request):
    ''' Django context processor to set 'browser' template variable.

        The 'browser' variable is a string containing common names of browsers
        compatible with the user's browser and platform. See
        http://web.koesbong.com/2011/01/28/python-css-browser-selector/

        Example

            <html class="{{ browser }}">
                ....
                {% if browser=='ie6' %}

                    ...

                {% endif %}
                ...

            </html>

        ---
    '''

    b = ' '.join(browser_types(request))
    #log('browser: %s' % b)
    return {'browser': b}

def css_ok(request):
    ''' Django context processor to set 'css_ok' template variable.

        The 'css_ok' variable is a boolean indicating if the user
        agent can properly display css.

        Example:

            {% if css_ok %}

                ...

            {% endif %}

    '''

    status = {'css_ok': not is_primitive_browser(request)}
    #log('css_ok: {}'.format(status))
    return status

def javascript_ok(request):
    ''' Django context processor to set 'javascript_ok' template variable.

        The 'javascript_ok' variable is a boolean indicating if the user agent
        can properly display javascript.

        Example

            {% if javascript_ok %}

                ...

            {% endif %}

        ---
    '''

    return {'javascript_ok': not is_primitive_browser(request)}


def is_live(request):
    ''' Django context processor to set 'is_live' template variable.

        The 'is_live' variable is a boolean indicating if the system is live.

        See debug()

        Example:

            {% if is_live %}

                ...

            {% endif %}

        '''
    try:
        live = settings.LIVE
    except Exception:
        live = False

    return {'is_live': live}


def debug(request):
    ''' Django context processor to set 'debug' template variable.

        The 'debug' variable is a boolean indicating if the system is in debug mode.

        The standard django.template.context_processors.debug sets 'debug',
        but only if:

        * in a RequestContext
        * settings.DEBUG is True
        * the request IP is in settings.INTERNAL_IPS

        This is too restrictive, especially the requirement for a RequestContext.

        Example:

            {% if debug %}

                ...

            {% endif %}

    '''

    return {'debug': settings.DEBUG}


def is_active_user(request):
    ''' Django context processor to set 'is_active_user' template variable.

        The 'debug' variable is a is_active_user indicating whether the use
        is authenticated and active.

        Example:

            {% if is_active_user %}

                ...


            {% endif %}

    '''

    try:
        active_user = {'is_active_user': request.user.is_authenticated and request.user.is_active}
    except Exception:
        active_user = {'is_active_user': False}

    return active_user



#! /usr/bin/python3
'''
    Python profiling.
    profile_addons is named to avoid conflict with python's profile pacakge.

    This module does not work well with threads. Run a profile inside the thread.
    Remember that profiling slows your code, so remove it when not needed.

    Copyright 2014-2020 DeNova
    Last modified: 2020-10-20

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import cProfile
import pstats
import time
from io import StringIO

from denova.python.log import get_log

log = get_log()

def run(command, datafile, globals=None, locals=None):
    ''' Profile python code

        'command' is python code as a string

        'datafile' is the pathname to write profile data.

        run() executes the python string in 'command'. Example:
            profile('myprogram.main()', '/tmp/main.profile')
        should be called in place of:
            myprogram.main()

        WARNING: stdout and stderr of the function profiles may go in the bit bucket (why?)
                 If you suspect an error in your code, run it without cProfile.

        >>> import os
        >>>
        >>> DATA = '/tmp/denova.python.profile.data'
        >>>
        >>> if os.path.exists(DATA):
        ...     os.remove(DATA)
        >>> run('_sample_test_code()', DATA, globals(), locals())
        start
        end
        >>> assert os.path.getsize(DATA)
    '''

    log.debug(f'run({repr(command)})')

    if globals is None and locals is None:
        cProfile.run(command, datafile)
    else:
        cProfile.runctx(command, globals, locals, filename=datafile, sort=-1)

    log.debug('run() done')

def report(datafile, lines=None):
    ''' Report on profile data from datafile.

        Default lines to print is 20.

        If you need to profile before the program ends, set a timer to
        invoke report().

        Returns report text.

        >>> import os, os.path
        >>>
        >>> DATA = '/tmp/denova.python.profile.data'
        >>> REPORT = '/tmp/denova.python.profile.report'
        >>>
        >>> for path in [DATA, REPORT]:
        ...     if os.path.exists(path):
        ...         os.remove(path)
        >>> run('_sample_test_code()', DATA, globals(), locals())
        start
        end
        >>> text = report(DATA)
        >>> assert text
        >>>
        >>> with open(REPORT, 'w') as f:
        ...     result = f.write(text)
        >>> os.path.getsize(REPORT) > 0
        True
    '''

    if not lines:
        lines = 20

    out = StringIO()
    stats = pstats.Stats(datafile, stream=out)
    # stats.strip_dirs()
    stats.sort_stats('cumulative', 'time', 'calls')
    stats.print_stats(lines)
    text = out.getvalue()

    log.debug(f'report from {datafile}:\n{text}')

    return text

def report_to_file(codestring, reportfile, datafile=None, globals=None, locals=None):
    ''' Profile codestring and write report to reportfile. '''

    def write_report():
        text = report(datafile)
        with open(reportfile, 'w') as f:
            f.write(text)
        log.debug(f'profile report is in {reportfile}')

    if datafile is None:
        datafile = reportfile + '.data'

    try:
        run(codestring, datafile, globals=globals, locals=locals)
    except:   # 'bare except' because it catches more than "except Exception"
        log.debug('always report profile')
        write_report()
        raise

    else:
        write_report()

def write_report(datafile, reportfile):
    ''' Write report from datafile to reportfile. '''

    text = report(datafile)
    with open(reportfile, 'w') as f:
        f.write(text)
    log.debug(f'profile report is in {reportfile}')

def _sample_test_code():
    ''' Sample code to use for profile testing. '''

    def start():
        print('start')

    def sleep():
        time.sleep(3)

    def end():
        print('end')

    start()
    sleep()
    end()


if __name__ == "__main__":

    import doctest
    doctest.testmod()



'''
    DeNova's open source libraries.

    All files in this directory tree are open source,
    licensed under GPLv3 <http://www.gnu.org/licenses/>.

    Copyright 2009-2020 DeNova
    Last modified: 2020-09-22
'''



#!/usr/bin/env python3
'''
    Command line interface.

    A wrapper for the pexpect module.
    The Responder class responds to prompts from programs.

    WARNING: All inputs to the cli module must be sanitised for security.

    Copyright 2013-2020 DeNova
    Last modified: 2020-10-20

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.

    import os
    import subprocess
    from pexpect import fdpexpect

    program = ['/path/to/command', '--arg1', 'value1', '--arg2', 'value2']
    devNull = open(os.devnull, 'w')
    command = subprocess.Popen(program, stdout=devNull,
                               stdin=subprocess.PIPE, stderr=subprocess.PIPE)
    child = fdpexpect.fdspawn(command.stderr)
    child.expect('hi')
'''
import os
import re

import pexpect

from denova.python.log import get_log

log = get_log()

class CliException(Exception):
    pass

class Responder():
    ''' Run a command line interface program with responses.

        'responses' is a list of::
            [(PROMPT, RESPONSE), ...]
        The PROMPT is a python regular expression so be careful
        to escape everything, as needed.

        >>> from denova.os.user import whoami
        >>> if whoami() == 'root':
        ...     responses = [
        ...        ('Are you sure you want to continue connecting \(yes/no\)\? ', 'n'),
        ...        ("Please type 'yes' or 'no':", 'no'),
        ...        ('$', 'exit'),
        ...      ]
        ...     responder = Responder(responses, 'ssh', 'localhost', _clilog=log).run()

        >>> if whoami() == 'root':
        ...     responses = [
        ...       ('localhost:.*?\$ ', 'ls'),
        ...       ('localhost:.*?\$ ', 'exit'),
        ...      ]
        ...     responder = Responder(responses, 'ssh', 'localhost').run()

        >>> if whoami() == 'root':
        ...     password = 'delete-this-user'
        ...     responses = [
        ...                  ('Enter new UNIX password: ', password),
        ...                  ('Retype new UNIX password: ', password),
        ...                  ('Full Name \[\]: ', None),
        ...                  ('Room Number \[]: ', None),
        ...                  ('Work Phone \[\]: ', None),
        ...                  ('Home Phone \[\]: ', None),
        ...                  ('Other \[\]: ', None),
        ...                  ('Is the information correct\? \[Y/n\] ', 'y'),
        ...                  ]
        ...     responder = Responder(responses, 'adduser', 'deleteme').run()
        ...     run('deluser', 'deleteme').returncode
        ... else:
        ...     print('0')
        0
    '''

    PRINT_LOG = False

    def __init__(self, responses, program, *args, **kwargs):
        self._clilog = None

        self.responses = responses
        self.command_line = f'{program}'

        for arg in args:
            self.command_line += ' '
            self.command_line += str(arg)
        for key, value in kwargs.items():
            self.command_line += ' '
            self.command_line += f'{key}={value}'
        log.debug(f'Responder command: "{self.command_line}"')

    def run(self):

        child = pexpect.spawn(self.command_line, encoding='utf-8', echo=False, logfile=log)

        log('responses:')
        for prompt, response in self.responses:
            log.debug(f'    response to {repr(prompt)} is {response}')
            child.expect(prompt)
            if response:
                child.sendline(response)
            else:
                log.warning('click return')
                child.sendline()

        try:
            child.interact()
        except pexpect.exceptions.TIMEOUT as pet:
            self._log(str(pet))
        except:   # 'bare except' because it catches more than "except Exception"
            self._log(str(child))
            raise

        child.close()
        self._log(f'exit status: {child.exitstatus}')
        self._log(f'signal status: {child.signalstatus}')

        return child.exitstatus

    def _log(self, line):
        ''' Log line. '''

        log.debug('output: {}'.format(line.strip('\n')))
        if self._clilog:
            self._clilog(line)
        elif self.PRINT_LOG:
            print('(no cli_log) ' + line.strip('\n'))


def minimal_env(user=None):
    '''
        Get very minimal, safe chroot env.

        Be sure to validate anything that comes from environment variables
        before using it. According to David A. Wheeler, a common cracker's
        technique is to change an environment variable.

        If user is not set, gets the user from denova.os.user.whoami(). This
        can flood /var/log/auth.log, so call with user set when you can.

        >>> from denova.os.user import whoami
        >>> env = minimal_env()
        >>> '/bin:/usr/bin:/usr/local/bin' in env['PATH']
        True
        >>> if whoami() == 'root':
        ...     '/sbin:/usr/sbin:/usr/local/sbin' in env['PATH']
        ... else:
        ...     '/sbin:/usr/sbin:/usr/local/sbin' not in env['PATH']
        True
    '''

    # import delayed to avoid recursive imports
    from denova.os.user import whoami

    if not user:
        user = whoami()

    env = {}

    # use a minimal path
    if user == 'root':
        env['PATH'] = '/bin:/usr/bin:/usr/local/bin:/sbin:/usr/sbin:/usr/local/sbin'
    else:
        env['PATH'] = '/bin:/usr/bin:/usr/local/bin'

    env_var = 'HOME'
    if env_var in os.environ:
        var = os.environ[env_var]
        # make sure the home directory is something reasonable and reasonably safe
        # Wheeler's Secure Programming warns against directories with '..'
        var = os.path.abspath(var)
        if os.path.exists(var):
            env[env_var] = var

    env_var = 'TZ'
    if env_var in os.environ:
        var = os.environ[env_var]
        # only set the variable if it's reasonable
        m = re.match('^([A-Za-z]+[A-Za-z_-]*/?[A-Za-z_-]*/?[A-Za-z_-]*?[A-Za-z0-9]*)$', var)
        if m and (m.group(1) == var):
            env[env_var] = var

    env_var = 'IFS'
    if env_var in os.environ:
        # force the variable to a known good value
        env[env_var] = "$' \t\n'"

    env_var = 'LC_ALL'
    if env_var in os.environ:
        # force the variable to a known good value
        env[env_var] = 'C'

    return env


if __name__ == "__main__":
    import doctest
    doctest.testmod()



'''
    Extra functions/templates, etc. when you have a django app.
    django_addons is named to avoid conflict with the django pacakge.

    Copyright 2009-2020 DeNova
    Last modified: 2020-10-02

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''



'''
    Iterator utilties.

    These classes are modeled after the ifunctions() in itertools and
    Dave Beasley's Generator Functions for System Programers.

    Portions Copyright 2011-2020 DeNova
    Last modified: 2020-10-10

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import os.path
import builtins
from itertools import takewhile
from tempfile import gettempdir

from denova.python.log import get_log

# log in non-standard dir because of concurrency issues
log = get_log(os.path.join(gettempdir(), 'iter.log'))

def is_iterable(item):
    ''' Return True iff item is iterable.

        >>> print(is_iterable(3))
        False
    '''

    try:
        list(item)
    except TypeError:
        result = False
    else:
        result = True

    return result

def iter(items, sentinel=None):
    ''' A more flexible and tolerant iter().
        Accepts a single item. Treats None as an empty iterable. '''

    if items is not None:
        try:
            if sentinel:
                items = takewhile(lambda x: not sentinel(x), items)
            for item in builtins.iter(items):
                yield item
        except TypeError:
            # does this make any sense?
            # does it belong above inside "if sentinel:"?
            # 'items' may a sentinel. If so, it is just one item.
            if sentinel:
                if not sentinel(items):
                    yield items
            else:
                yield items

def count(iterable):
    ''' Count items in any iterable, or an object which can be made into an
        iterable.

        This is what you probably expected itertools.count() to do.

        Since count() works with any iterable, you don't have to remember
        when you can use len(). It can also be much more memory efficient
        than len(list(iterator)), since like generators it only requires
        one item in memory at a time.

        Traverses the iterable. This can result in side effects, such as
        exhausting a generator.

        If the iterable is a sequence (string, tuple or list) or a
        mapping (dictionary), the built-in function len() does not have side
        effects and may be more efficient. '''

    return sum(1 for item in iter(iterable))

def first(test, items):
    ''' Return first item for which test(item) is True. '''

    matches = list(filter(test, items))
    if matches:
        result = matches[0]
    else:
        result = None

    return result

def first_true(items):
    ''' Return first item in iterable that evaluates to True.
        If all items are False, return None.

        >>> print(first_true([0, None, '', 'yes', 'yes again', False]))
        yes
        >>> print(first_true([0, None, '', 'yes again', 'yes', False]))
        yes again
        >>> print(first_true([0, None, '', False]))
        None
    '''

    try:
        iterable = iter(items)
        item = next(iterable)
        while not item:
            item = next(iterable)
    except StopIteration:
        item = None
    return item

def last_true(iterable):
    ''' Return last item in iterable that evaluates to True.
        If all items are False, return None.

        Uses minimum memory, unlike "reverse(list(iterable))[0]".

        >>> print(last_true([0, None, '', 'yes', 'yes again', False]))
        yes again
    '''

    last_true_item = None
    for item in iterable:
        if item:
            last_true_item = item
    return last_true_item

def enqueue(q, *iterables):
    ''' Queue items from one or more iterables.

        You can enqueue() items in one thread, and dequeue() them from another.
        As of Python 2.5 Queue.task_done() and join() seem buggy, so use
        another way to let the dequeue thread run, such as threading.Event or
        if necessary time.sleep(). Set up the thread that calls dequeue(),
        then call enqueue(). See the doctest example.

        StopIteration can not be an item from one of the iterables, which is
        vanishingly unlikely anyway.

        Thanks to Dave Beasley.

        >>> from queue import Queue

        >>> iter1 = (1, 2)
        >>> iter2 = [3, 4]
        >>> q = Queue(0)

        >>> enqueue(q, iter1, iter2)
        >>> for item in dequeue(q):
        ...     print(item)
        1
        2
        3
        4

        >>> def dequeue_test(q, done_event):
        ...     for item in dequeue(q):
        ...         print(item)
        ...     done_event.set()
        >>> from threading import Event, Thread
        >>> done_event = Event()
        >>> thread = Thread(target=dequeue_test, args=(q, done_event))
        >>> thread.start()
        >>> enqueue(q, iter1, iter2)
        >>> done_event.wait()
        1
        2
        3
        4
        True
        '''

    for iterable in iterables:
        for item in iterable:
            q.put(item)
    # will this work across net links? python instances on a machine?
    q.put(StopIteration)

def dequeue(q):
    ''' Get items that were queued using enqueue().

        You can enqueue() items in one thread, and dequeue() them from another.

        Thanks to Dave Beasley. '''

    item = q.get()
    while item != StopIteration:
        yield item
        item = q.get()


if __name__ == "__main__":

    import doctest
    doctest.testmod()
